# Copyright (c) 2011, Dirk Thomas, Dorian Scholz, TU Darmstadt
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above
#     copyright notice, this list of conditions and the following
#     disclaimer in the documentation and/or other materials provided
#     with the distribution.
#   * Neither the name of the TU Darmstadt nor the names of its
#     contributors may be used to endorse or promote products derived
#     from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from python_qt_binding.QtCore import qDebug, QRect, Qt, Signal
from python_qt_binding.QtGui import QFont
from python_qt_binding.QtWidgets import QTextBrowser, QToolBar

from qt_gui.dockable_main_window import DockableMainWindow
from qt_gui.settings import Settings


class MainWindow(DockableMainWindow):
    """Main window of the application managing the geometry and state of all top-level widgets."""

    save_settings_before_close_signal = Signal(Settings, Settings)

    def __init__(self, help_text=None):
        super(MainWindow, self).__init__()
        self.setObjectName('MainWindow')
        self._help_widget = None

        if help_text:
            font = QFont()
            font.setPointSize(14)

            self._help_widget = QTextBrowser(self)
            self._help_widget.setFont(font)
            self._help_widget.setReadOnly(True)
            self._help_widget.setTextInteractionFlags(Qt.TextBrowserInteraction)
            self._help_widget.setOpenExternalLinks(True)
            self._help_widget.setHtml(help_text)
            self._help_widget.setStyleSheet('background:transparent;')

        self._save_on_close_signaled = False
        self._global_settings = None
        self._perspective_settings = None
        self._settings = None

    def showHelpWidget(self, should_show):
        if self._help_widget:
            self._help_widget.show() if should_show else self._help_widget.hide()

    def resizeEvent(self, event):
        if self._help_widget:
            size = self.size()
            info_percentage = 0.75
            margin_percentage = (1.0 - info_percentage) / 2
            self._help_widget.setGeometry(
                QRect(int(size.width() * margin_percentage),
                      int(size.height() * margin_percentage),
                      int(size.width() * info_percentage),
                      int(size.height() * info_percentage)))

    def closeEvent(self, event):
        qDebug('MainWindow.closeEvent()')
        if not self._save_on_close_signaled:
            self._save_geometry_to_perspective()
            self._save_state_to_perspective()
            self._save_on_close_signaled = True
            self.save_settings_before_close_signal.emit(
                self._global_settings, self._perspective_settings)
            event.ignore()
        else:
            event.accept()

    def save_settings(self, global_settings, perspective_settings):
        qDebug('MainWindow.save_settings()')
        self._global_settings = global_settings
        self._perspective_settings = perspective_settings
        self._settings = self._perspective_settings.get_settings('mainwindow')
        # store current setup to current _settings
        self._save_geometry_to_perspective()
        self._save_state_to_perspective()

    def restore_settings(self, global_settings, perspective_settings):
        qDebug('MainWindow.restore_settings()')
        # remember new _settings
        self._global_settings = global_settings
        self._perspective_settings = perspective_settings
        self._settings = self._perspective_settings.get_settings('mainwindow')
        # only restore geometry, restoring state is triggered after PluginManager has been updated
        self._restore_geometry_from_perspective()

    def save_setup(self):
        qDebug('MainWindow.save_setup()')
        # store current setup to current _settings
        self._save_geometry_to_perspective()
        self._save_state_to_perspective()

    def restore_state(self):
        qDebug('MainWindow.restore_state()')
        # restore state from _settings
        self._restore_state_from_perspective()

    def perspective_changed(self, name):
        self.setWindowTitle('%s - rqt' % str(name))

    def _save_geometry_to_perspective(self):
        if self._settings is not None:
            # unmaximizing widget before saveGeometry works around bug to restore dock-widgets
            # still the non-maximized size can not correctly be restored
            maximized = self.isMaximized()
            if maximized:
                self.showNormal()
            self._settings.set_value('geometry', self.saveGeometry())
            if maximized:
                self.showMaximized()

    def _restore_geometry_from_perspective(self):
        if self._settings.contains('geometry'):
            self.restoreGeometry(self._settings.value('geometry'))

    def _save_state_to_perspective(self):
        if self._settings is not None:
            self._settings.set_value('state', self.saveState())
            # safe area for all toolbars
            toolbar_settings = self._settings.get_settings('toolbar_areas')
            for toolbar in self.findChildren(QToolBar):
                area = self.toolBarArea(toolbar)
                if area in [Qt.LeftToolBarArea,
                            Qt.RightToolBarArea,
                            Qt.TopToolBarArea,
                            Qt.BottomToolBarArea]:
                    toolbar_settings.set_value(toolbar.objectName(), area)

    def _restore_state_from_perspective(self):
        if self._settings.contains('state'):
            self.restoreState(self._settings.value('state'))
            # restore area for all toolbars
            toolbar_settings = self._settings.get_settings('toolbar_areas')
            for toolbar in self.findChildren(QToolBar):
                if not toolbar.objectName():
                    continue
                area = Qt.ToolBarArea(
                    int(toolbar_settings.value(toolbar.objectName(), Qt.NoToolBarArea)))
                if area in [Qt.LeftToolBarArea,
                            Qt.RightToolBarArea,
                            Qt.TopToolBarArea,
                            Qt.BottomToolBarArea]:
                    self.addToolBar(area, toolbar)



# Copyright (c) 2011, Dirk Thomas, Dorian Scholz, TU Darmstadt
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above
#     copyright notice, this list of conditions and the following
#     disclaimer in the documentation and/or other materials provided
#     with the distribution.
#   * Neither the name of the TU Darmstadt nor the names of its
#     contributors may be used to endorse or promote products derived
#     from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import os
import platform
import sys

from python_qt_binding import QT_BINDING, QT_BINDING_MODULES, QT_BINDING_VERSION
from python_qt_binding.QtCore import QObject, qVersion
from python_qt_binding.QtWidgets import QMessageBox

from .ros_package_helper import get_package_path


class AboutHandler(QObject):
    """
    Handler for the 'About action' menu bar item.

    This handler shows a message box with details on the used libraries and their versions.
    """

    def __init__(self, qtgui_path, parent=None):
        super(AboutHandler, self).__init__(parent)
        self._qtgui_path = qtgui_path

    def show(self):
        try:
            # append folder of 'qt_gui_cpp/lib' to module search path
            qt_gui_cpp_path = os.path.realpath(get_package_path('qt_gui_cpp'))
        except Exception:
            qt_gui_cpp = None
        else:
            sys.path.append(os.path.join(qt_gui_cpp_path, 'lib'))
            sys.path.append(os.path.join(qt_gui_cpp_path, 'src'))
            from qt_gui_cpp.cpp_binding_helper import qt_gui_cpp

        logo = os.path.join(
            self._qtgui_path, 'share', 'qt_gui', 'resource', 'ros_org_vertical.png')
        text = '<img src="%s" width="56" height="200" style="float: left;"/>' % logo

        text += '<h3 style="margin-top: 1px;">%s</h3>' % self.tr('rqt')

        text += '<p>%s %s</p>' % (
            self.tr('rqt is a framework for graphical user interfaces.'),
            self.tr('It is extensible with plugins which can be written in either Python or C++.'))
        text += '<p>%s</p>' % (
            self.tr(
                'Please see the <a href="%s">Wiki</a> for more information on rqt and available '
                'plugins.' % 'http://wiki.ros.org/rqt'))

        text += '<p>%s: ' % self.tr('Utilized libraries:')

        text += 'Python %s, ' % platform.python_version()

        if QT_BINDING == 'pyside':
            text += 'PySide'
        elif QT_BINDING == 'pyqt':
            text += 'PyQt'
        text += ' %s (%s), ' % (QT_BINDING_VERSION, ', '.join(sorted(QT_BINDING_MODULES)))

        text += 'Qt %s, ' % qVersion()

        if qt_gui_cpp is not None:
            if QT_BINDING == 'pyside':
                text += '%s' % (self.tr('%s C++ bindings available') % 'Shiboken')
            elif QT_BINDING == 'pyqt':
                text += '%s' % (self.tr('%s C++ bindings available') % 'SIP')
            else:
                text += '%s' % self.tr('C++ bindings available')
        else:
            text += '%s' % self.tr('no C++ bindings found - no C++ plugins available')

        text += '.</p>'

        mb = QMessageBox(
            QMessageBox.NoIcon, self.tr('About rqt'), text, QMessageBox.Ok, self.parent())
        mb.exec_()



# Copyright (c) 2011, Dirk Thomas, Dorian Scholz, TU Darmstadt
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above
#     copyright notice, this list of conditions and the following
#     disclaimer in the documentation and/or other materials provided
#     with the distribution.
#   * Neither the name of the TU Darmstadt nor the names of its
#     contributors may be used to endorse or promote products derived
#     from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from python_qt_binding.QtCore import QObject


class PluginContext(QObject):
    """
    Provides information to the plugin and exposes methods to interact with the framework.

    PluginContext relays all methods to the corresponding `PluginHandler`.
    """

    def __init__(self, handler):
        super(PluginContext, self).__init__(handler)
        self.setObjectName('PluginContext')

        self._handler = handler

    def serial_number(self):
        """
        Return the serial number of the plugin.

        For a specific type of plugin each instance gets a serial number (which is the first
        currently not used positive integer at construction time).
        @return: The serial number
        @rtype: int
        """
        return self._handler.instance_id().serial_number

    def argv(self):
        """
        Return the command line arguments of the plugin.

        @return: The arguments without a program name at the beginning
        @rtype: list
        """
        return self._handler.argv()

    def add_widget(self, widget):
        """
        Add a widget to the UI.

        The widget is embedded into a new QDockWidget which itself is added to the QMainWindow.
        This method can be called once for each widget a plugin would like to add and at any point
        in time (until the calling plugin has been shutdown).

        Note: The ownership of the widget is transferred to the callee which will delete it when
            the plugin is shut down.
        @param widget: The widget to add
        @type widget: QWidget
        """
        self._handler.add_widget(widget)

    def remove_widget(self, widget):
        """
        Remove a previously added widget from the UI.

        Note: The ownership of the widget is transferred to the caller.
        @param widget: The widget to remove
        @type widget: QWidget
        """
        self._handler.remove_widget(widget)

    def add_toolbar(self, toolbar):
        """
        Add a toolbar to the UI, which is directly added to the QMainWindow.

        This method can be called once for each toolbar a plugin would like to add and at any point
        in time (until the calling plugin has been shutdown).
        Note: The ownership of the toolbar is transferred to the callee which will delete it when
              the plugin is shut down.
        @param widget: The toolbar to add
        @type widget: QToolBar
        """
        self._handler.add_toolbar(toolbar)

    def remove_toolbar(self, toolbar):
        """
        Remove a previously added toolbar from the UI.

        Note: The ownership of the toolbar is transferred to the caller.
        @param widget: The toolbar to remove
        @type widget: QToolBar
        """
        self._handler.remove_toolbar(toolbar)

    def close_plugin(self):
        """
        Close the plugin.

        The framework will call `Plugin.shutdown_plugin()` and unload it afterwards.
        """
        self._handler.close_plugin()

    def reload_plugin(self):
        """Reload the plugin."""
        self._handler.reload_plugin()



# Copyright (c) 2011, Dirk Thomas, Dorian Scholz, TU Darmstadt
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above
#     copyright notice, this list of conditions and the following
#     disclaimer in the documentation and/or other materials provided
#     with the distribution.
#   * Neither the name of the TU Darmstadt nor the names of its
#     contributors may be used to endorse or promote products derived
#     from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import traceback

from python_qt_binding.QtCore import qCritical

from qt_gui.plugin_provider import PluginProvider


class CompositePluginProvider(PluginProvider):
    """Composite of multiple `PluginProvider`s."""

    def __init__(self, plugin_providers=None):
        super(CompositePluginProvider, self).__init__()
        self.setObjectName('CompositePluginProvider')

        self._plugin_providers = plugin_providers or []
        self._discovered_plugins = {}
        self._running_plugins = {}

    def set_plugin_providers(self, plugin_providers):
        self._plugin_providers = plugin_providers

    def discover(self, discovery_data):
        # discover plugins from all providers
        discovered_plugins = []
        for plugin_provider in self._plugin_providers:
            try:
                plugin_descriptors = plugin_provider.discover(discovery_data)
            except Exception:
                qCritical(
                    'CompositePluginProvider.discover() could not discover plugins from '
                    'provider "%s":\n%s' % (type(plugin_provider), traceback.format_exc()))
            else:
                self._discovered_plugins[plugin_provider] = plugin_descriptors
                discovered_plugins += plugin_descriptors
        return discovered_plugins

    def load(self, plugin_id, plugin_context):
        # dispatch load to appropriate provider
        for plugin_provider, plugin_descriptors in self._discovered_plugins.items():
            for plugin_descriptor in plugin_descriptors:
                if plugin_descriptor.plugin_id() == plugin_id:
                    instance = plugin_provider.load(plugin_id, plugin_context)
                    self._running_plugins[instance] = plugin_provider
                    return instance
        raise UserWarning('plugin_id "%s" not found' % plugin_id)

    def unload(self, plugin_instance):
        # dispatch unload to appropriate provider
        if plugin_instance in self._running_plugins:
            self._running_plugins[plugin_instance].unload(plugin_instance)
            self._running_plugins.pop(plugin_instance)
            return
        raise UserWarning('plugin_instance not found')

    def shutdown(self):
        for plugin_provider in self._plugin_providers:
            plugin_provider.shutdown()



# Copyright (c) 2011, Dirk Thomas, Dorian Scholz, TU Darmstadt
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above
#     copyright notice, this list of conditions and the following
#     disclaimer in the documentation and/or other materials provided
#     with the distribution.
#   * Neither the name of the TU Darmstadt nor the names of its
#     contributors may be used to endorse or promote products derived
#     from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from python_qt_binding.QtGui import QIcon

from qt_gui.dock_widget import DockWidget
from qt_gui.dockable_main_window import DockableMainWindow


class DockWidgetContainer(DockWidget):
    """`DockWidget` containing a main window acting as a container for other dock widgets."""

    def __init__(self, container_manager, serial_number):
        super(DockWidgetContainer, self).__init__(container_manager)
        self._serial_number = serial_number
        self._settings = None

        self.main_window = DockableMainWindow()
        self.setWidget(self.main_window)
        self.setWindowIcon(QIcon.fromTheme('folder'))

    def serial_number(self):
        return self._serial_number

    def setObjectName(self, name):
        super(DockWidget, self).setObjectName(name)
        self.main_window.setObjectName(name + '__MainWindow')

    def save_settings(self, settings):
        mw_settings = settings.get_settings('mainwindow')
        self._save_geometry(mw_settings)
        self._save_state(mw_settings)
        super(DockWidgetContainer, self).save_settings(settings)

    def _save_geometry(self, settings):
        # unmaximizing widget before saveGeometry works around bug to restore dock-widgets
        # still the non-maximized size can not correctly be restored
        maximized = self.isMaximized()
        if maximized:
            self.showNormal()
        settings.set_value('geometry', self.main_window.saveGeometry())
        if maximized:
            self.showMaximized()

    def _save_state(self, settings):
        if self._settings is not None:
            self._settings.set_value('state', self.main_window.saveState())

    def restore_settings(self, settings):
        super(DockWidgetContainer, self).restore_settings(settings)
        mw_settings = settings.get_settings('mainwindow')
        self._settings = mw_settings
        # only restore geometry, restoring state is triggered after PluginManager has been updated
        self._restore_geometry(mw_settings)

    def _restore_geometry(self, settings):
        if settings.contains('geometry'):
            self.main_window.restoreGeometry(settings.value('geometry'))

    def restore_state(self):
        if self._settings.contains('state'):
            self.main_window.restoreState(self._settings.value('state'))



# Copyright (c) 2011, Dirk Thomas, Dorian Scholz, TU Darmstadt
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above
#     copyright notice, this list of conditions and the following
#     disclaimer in the documentation and/or other materials provided
#     with the distribution.
#   * Neither the name of the TU Darmstadt nor the names of its
#     contributors may be used to endorse or promote products derived
#     from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import os

from python_qt_binding import loadUi
from python_qt_binding.QtCore import QEvent, QObject, Qt, qWarning
from python_qt_binding.QtGui import QIcon
from python_qt_binding.QtWidgets import QDockWidget, QMenu, QWidget


class DockWidgetTitleBar(QWidget):
    """Title bar for dock widgets providing custom actions."""

    def __init__(self, dock_widget, qtgui_path, hide_title=False):
        super(DockWidgetTitleBar, self).__init__(dock_widget)
        self._dock_widget = dock_widget

        ui_file = os.path.join(
            qtgui_path, 'share', 'qt_gui', 'resource', 'dock_widget_title_bar.ui')
        loadUi(ui_file, self)
        self._extra_buttons = {
            'configuration': self.configuration_button,
            'reload': self.reload_button,
            'help': self.help_button,
            'close': self.close_button,
        }

        icon = QIcon.fromTheme('emblem-system')
        if not icon.isNull():
            self.configuration_button.setIcon(icon)
            self.configuration_button.setText('')
        icon = QIcon.fromTheme('view-refresh')
        if not icon.isNull():
            self.reload_button.setIcon(icon)
            self.reload_button.setText('')
        icon = QIcon.fromTheme('help-browser')
        if not icon.isNull():
            self.help_button.setIcon(icon)
            self.help_button.setText('')

        icon = QIcon.fromTheme('window-close')
        if not icon.isNull():
            self.close_button.setIcon(icon)
            self.close_button.setText('')
        self.close_button.clicked.connect(self._close_clicked)

        self.float_button.clicked.connect(self._toggle_floating)
        self.dockable_button.clicked[bool].connect(self._toggle_dockable)
        self.minimize_button.clicked.connect(self._minimize_dock_widget)

        self._dock_widget.featuresChanged.connect(self._features_changed)
        self._features_changed()

        self._update_icon()
        self._update_title()

        self._close_callbacks = []
        self._event_callbacks = {
            QEvent.WindowIconChange: self._update_icon,
            QEvent.WindowTitleChange: self._update_title,
        }
        self._dock_widget.installEventFilter(self)

        self.title_label.installEventFilter(self)
        self.title_edit.hide()
        self.title_edit.editingFinished.connect(self._finished_editing)
        self.title_edit.returnPressed.connect(self._update_title_label)

        if hide_title:
            # hide all except the configuration button if present
            self.icon_label.hide()
            self.title_label.hide()
            self.dockable_button.hide()
            self.reload_button.hide()
            self.help_button.hide()
            self.minimize_button.hide()
            self.float_button.hide()
            self.close_button.hide()

    def connect_button(self, button_id, callback):
        button = self._extra_buttons.get(button_id, None)
        if button is None:
            qWarning('DockWidgetTitleBar.connect_button(): unknown button_id: %s' % button_id)
            return
        button.clicked.connect(callback)

    def connect_close_button(self, callback):
        self._close_callbacks.append(callback)

    def _close_clicked(self):
        for callback in self._close_callbacks:
            callback(self.parent())

    def show_button(self, button_id, visibility=True):
        button = self._extra_buttons.get(button_id, None)
        if button is None:
            qWarning('DockWidgetTitleBar.show_button(): unknown button_id: %s' % button_id)
            return
        button.setVisible(visibility)

    def hide_button(self, button_id):
        self.show_button(button_id, False)

    def eventFilter(self, obj, event):
        if event.type() in self._event_callbacks:
            ret_val = self._event_callbacks[event.type()](obj, event)
            if ret_val is not None:
                return ret_val
        if event.type() == event.ContextMenu and obj == self.title_label:
            menu = QMenu(self)
            rename_action = menu.addAction(self.tr('Rename dock widget'))
            action = menu.exec_(self.mapToGlobal(event.pos()))
            if action == rename_action:
                self.title_label.hide()
                self.title_edit.setText(self.title_label.text())
                self.title_edit.show()
                self.title_edit.setFocus()
            return True
        return QObject.eventFilter(self, obj, event)

    def _update_icon(self, *args):
        pixmap = None
        if self.parentWidget().windowIcon():
            pixmap = self.parentWidget().windowIcon().pixmap(self.close_button.iconSize())
        self.icon_label.setPixmap(pixmap)

    def _update_title(self, *args):
        self.title_label.setText(self.parentWidget().windowTitle())

    def _toggle_dockable(self, enabled):
        dock_widget = self.parentWidget()
        if enabled:
            dock_widget.setAllowedAreas(Qt.AllDockWidgetAreas)
        else:
            dock_widget.setAllowedAreas(Qt.NoDockWidgetArea)

    def _toggle_floating(self):
        dock_widget = self.parentWidget()
        dock_widget.setFloating(not dock_widget.isFloating())

    def _minimize_dock_widget(self):
        dock_widget = self.parentWidget()
        dock_widget.hide()

    def _features_changed(self, features=None):
        if features is None:
            features = self.parentWidget().features()

        closable = bool(features & QDockWidget.DockWidgetClosable)
        self.close_button.setVisible(closable)
        self.reload_button.setVisible(closable)

        movable = bool(features & QDockWidget.DockWidgetMovable)
        self.dockable_button.setChecked(movable)
        self._toggle_dockable(self.dockable_button.isChecked())
        self.dockable_button.setVisible(movable)
        self.float_button.setVisible(movable)
        self.minimize_button.setVisible(movable)

    def save_settings(self, settings):
        settings.set_value('dock_widget_title', self._dock_widget.windowTitle())

        # skip saving dockable flag when layout is frozen
        movable = bool(self.parentWidget().features() & QDockWidget.DockWidgetMovable)
        if movable:
            settings.set_value('dockable', self.dockable_button.isChecked())

    def restore_settings(self, settings):
        dock_widget_title = settings.value('dock_widget_title', None)
        if dock_widget_title is not None:
            self.title_label.setText(dock_widget_title)
            self._dock_widget.setWindowTitle(dock_widget_title)

        dockable = settings.value('dockable', True) in [True, 'true']
        # only allow dockable when layout is not frozen
        movable = bool(self.parentWidget().features() & QDockWidget.DockWidgetMovable)
        self.dockable_button.setChecked(dockable and movable)
        self._toggle_dockable(self.dockable_button.isChecked())

    def _finished_editing(self):
        self.title_edit.hide()
        self.title_label.show()

    def _update_title_label(self):
        if self.title_edit.text():
            self.title_label.setText(self.title_edit.text())
            self._dock_widget.setWindowTitle(self.title_edit.text())


if __name__ == '__main__':
    import sys
    from python_qt_binding.QtWidgets import QApplication
    from qt_gui.dockable_main_window import DockableMainWindow
    from ament_index_python.resources import get_resource

    _, qtgui_path = get_resource('packages', 'qt_gui')
    app = QApplication(sys.argv)

    win = DockableMainWindow()

    dock1 = QDockWidget('dockwidget1', win)
    win.addDockWidget(Qt.LeftDockWidgetArea, dock1)
    title_bar = DockWidgetTitleBar(dock1, qtgui_path)
    dock1.setTitleBarWidget(title_bar)

    dock2 = QDockWidget('dockwidget2')
    win.addDockWidget(Qt.RightDockWidgetArea, dock2)
    title_bar = DockWidgetTitleBar(dock2, qtgui_path)

    dock2.setTitleBarWidget(title_bar)

    win.resize(640, 480)
    win.show()

    app.exec_()



# Software License Agreement (BSD License)
#
# Copyright (c) 2013, Open Source Robotics Foundation, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above
#    copyright notice, this list of conditions and the following
#    disclaimer in the documentation and/or other materials provided
#    with the distribution.
#  * Neither the name of Open Source Robotics Foundation, Inc. nor
#    the names of its contributors may be used to endorse or promote
#    products derived from this software without specific prior
#    written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import os

from python_qt_binding.QtGui import QIcon


def get_icon(name, type_=None, base_path=None):
    if type_ == 'file' or type_ is None:
        path = name
        if base_path is not None:
            path = os.path.join(base_path, path)
        icon = QIcon(path)
        if len(icon.availableSizes()) == 0:
            raise UserWarning('icon "%s" not found' % str(path))
    elif type_ == 'resource':
        icon = QIcon(name)
        if len(icon.availableSizes()) == 0:
            raise UserWarning('icon "%s" not found' % str(path))
    elif type_ == 'theme':
        # see http://standards.freedesktop.org/icon-naming-spec/icon-naming-spec-latest.html
        icon = QIcon.fromTheme(name)
    else:
        raise UserWarning('unknown icon type "%s"' % str(type_))
    return icon



# Software License Agreement (BSD License)
#
# Copyright (c) 2012, Willow Garage, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above
#    copyright notice, this list of conditions and the following
#    disclaimer in the documentation and/or other materials provided
#    with the distribution.
#  * Neither the name of Willow Garage, Inc. nor the names of its
#    contributors may be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import __builtin__
from importlib import reload
import os
import sys


class ReloadImporter:
    """
    Overrides the builtin import and automatically reloads all modules.

    Modules are imported from one of the reload paths after calling enable.
    """

    def __init__(self):
        self._excluded_modules = sys.modules.keys()
        self._reload_paths = None
        self._import_stack = []
        self._reloaded_modules = set()
        self._import = __builtin__.__import__

    def enable(self):
        __builtin__.__import__ = self._reimport

    def disable(self):
        __builtin__.__import__ = self._import

    def add_reload_path(self, path):
        if self._reload_paths is None:
            self._reload_paths = ()
        self._reload_paths += (os.path.abspath(path),)

    def _reload(self, module):
        if module.__name__ not in self._import_stack and module.__name__ in sys.modules:
            if not self._import_stack:
                self._reloaded_modules.clear()
            self._import_stack.append(module.__name__)
            # force reload
            if module.__name__ not in self._reloaded_modules:
                reload(module)
                self._reloaded_modules.add(module.__name__)
            self._import_stack.pop()

    def _reimport(self, name, globals_=None, locals_=None, fromlist=None, level=-1):
        module = self._import(
            name, globals_, locals_, fromlist if not None else [], level if not None else -1)

        if module.__name__ not in self._excluded_modules:
            if self._reload_paths is None or \
                    (hasattr(module, '__file__') and
                     len([p for p in self._reload_paths if module.__file__.startswith(p)]) > 0):
                self._reload(module)

        return module



# Copyright (c) 2011, Dirk Thomas, Dorian Scholz, TU Darmstadt
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above
#     copyright notice, this list of conditions and the following
#     disclaimer in the documentation and/or other materials provided
#     with the distribution.
#   * Neither the name of the TU Darmstadt nor the names of its
#     contributors may be used to endorse or promote products derived
#     from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.


class PluginInstanceId():
    """Identifier of a plugin instance."""

    def __init__(self, plugin_id=None, serial_number=None, instance_id=None):
        if instance_id is not None:
            # convert from unicode
            instance_id = str(instance_id)
            parts = instance_id.rsplit('#', 1)
            self.plugin_id = parts[0]
            self.serial_number = int(parts[1])
        else:
            # convert from unicode
            self.plugin_id = str(plugin_id)
            self.serial_number = int(serial_number) if serial_number is not None else None

    def __eq__(self, other):
        return self.plugin_id == other.plugin_id and self.serial_number == other.serial_number

    def __hash__(self):
        return hash(str(self))

    def __str__(self):
        return self.plugin_id + '#' + str(self.serial_number)

    def tidy_str(self):
        return self.plugin_id.replace('/', '__') + '__' + str(self.serial_number)

    def tidy_plugin_str(self):
        return self.plugin_id.replace('/', '__')



# Software License Agreement (BSD License)
#
# Copyright (c) 2012, Willow Garage, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above
#    copyright notice, this list of conditions and the following
#    disclaimer in the documentation and/or other materials provided
#    with the distribution.
#  * Neither the name of Willow Garage, Inc. nor the names of its
#    contributors may be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from python_qt_binding.QtCore import QEvent, QObject, Signal


class WindowChangedSignaler(QObject):
    """Signaler for for filtered events of an icon or title change of a widget."""

    hide_signal = Signal(object)
    show_signal = Signal(object)

    window_icon_changed_signal = Signal(object)
    window_title_changed_signal = Signal(object)

    def __init__(self, widget, parent=None):
        super(WindowChangedSignaler, self).__init__(parent)
        self._widget = widget
        self.setObjectName('WindowChangedSignaler__' + self._widget.objectName())
        self._recursive_invocation = False
        self._widget.installEventFilter(self)

    def emit_all(self):
        if self._widget.isVisible():
            self.show_signal.emit(self._widget)
        else:
            self.hide_signal.emit(self._widget)
        self.window_icon_changed_signal.emit(self._widget)
        self.window_title_changed_signal.emit(self._widget)

    def eventFilter(self, watched, event):
        if event.type() == QEvent.Hide:
            self.hide_signal.emit(watched)

        if event.type() == QEvent.Show:
            self.show_signal.emit(watched)

        if event.type() == QEvent.WindowIconChange:
            # prevent emitting recursive event when emitted signal will trigger another event
            if not self._recursive_invocation:
                self._recursive_invocation = True
                self.window_icon_changed_signal.emit(watched)
                self._recursive_invocation = False

        if event.type() == QEvent.WindowTitleChange:
            # prevent emitting recursive event when emitted signal will trigger another event
            if not self._recursive_invocation:
                self._recursive_invocation = True
                self.window_title_changed_signal.emit(watched)
                self._recursive_invocation = False

        return QObject.eventFilter(self, watched, event)



# Copyright (c) 2011, Dirk Thomas, Dorian Scholz, TU Darmstadt
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above
#     copyright notice, this list of conditions and the following
#     disclaimer in the documentation and/or other materials provided
#     with the distribution.
#   * Neither the name of the TU Darmstadt nor the names of its
#     contributors may be used to endorse or promote products derived
#     from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import time
import traceback

from python_qt_binding.QtCore import \
    qCritical, qDebug, QObject, Qt, qWarning, Signal, Slot

from qt_gui.errors import PluginLoadError
from qt_gui.plugin_handler_container import PluginHandlerContainer
from qt_gui.plugin_handler_direct import PluginHandlerDirect
from qt_gui.plugin_instance_id import PluginInstanceId
from qt_gui.plugin_menu import PluginMenu
from qt_gui.settings import Settings
from qt_gui.settings_proxy import SettingsProxy


class PluginManager(QObject):
    """
    Manager of plugin life cycle.

    It creates a specific `PluginHandler` for each plugin instance and maintains the perspective
    specific set of running plugins.
    """

    plugin_load_unload_signal = Signal(int)
    plugins_about_to_change_signal = Signal()
    plugins_changed_signal = Signal()
    plugin_help_signal = Signal(object)
    save_settings_completed_signal = Signal()
    close_application_signal = Signal()
    _deferred_reload_plugin_signal = Signal(str)

    discovery_cache_max_age = 60 * 60 * 24  # one day

    def __init__(self, plugin_provider, settings, application_context, settings_prefix=''):
        super(PluginManager, self).__init__()
        self.setObjectName('PluginManager')

        self._plugin_provider = plugin_provider
        self._settings = Settings(SettingsProxy(settings), '/'.join(
            [x for x in ['plugin_manager', settings_prefix] if x != '']))
        self._application_context = application_context

        self._main_window = None
        self._container_manager = None
        self._plugin_menu = None
        self._minimized_dock_widgets_toolbar = None

        self._global_settings = None
        self._perspective_settings = None
        self._plugin_descriptors = None
        self._running_plugins = {}

        self._number_of_ongoing_calls = None

        if self._application_context.options.multi_process or \
                self._application_context.options.embed_plugin:
            try:
                from qt_gui.plugin_handler_xembed import PluginHandlerXEmbed  # noqa: F401
            except ImportError:
                qCritical('PluginManager.__init__() multiprocess-mode only available under linux')
                exit(-1)

        # force connection type to queued, to delay the 'reloading' giving the
        # 'unloading' time to finish
        self._deferred_reload_plugin_signal.connect(
            self._reload_plugin_load, type=Qt.QueuedConnection)

        if self._application_context.provide_app_dbus_interfaces:
            from qt_gui.plugin_manager_dbus_interface import PluginManagerDBusInterface
            self._dbus_service = PluginManagerDBusInterface(self, self._application_context)

    def set_main_window(self, main_window, menu_bar, container_manager):
        self._main_window = main_window
        self._container_manager = container_manager
        self.plugins_changed_signal.connect(self._container_manager.restore_state_of_containers)
        if menu_bar is not None:
            self._plugin_menu = PluginMenu(menu_bar, self)
            self._plugin_menu.load_plugin_signal.connect(self.load_plugin)
            self._plugin_menu.unload_plugin_signal.connect(self.unload_plugin)

    def set_minimized_dock_widgets_toolbar(self, toolbar):
        self._minimized_dock_widgets_toolbar = toolbar

    def discover(self):
        # skip discover if called multiple times
        if self._plugin_descriptors is not None:
            return
        self._plugin_descriptors = {}

        # register discovered plugins
        plugin_descriptors = self._discover()
        for plugin_descriptor in plugin_descriptors:
            self._plugin_descriptors[plugin_descriptor.plugin_id()] = plugin_descriptor

            if self._plugin_menu is not None:
                self._plugin_menu.add_plugin(plugin_descriptor)

        if self._container_manager is not None:
            descriptor = self._container_manager.get_container_descriptor()
            self._plugin_descriptors[descriptor.plugin_id()] = descriptor
            if self._plugin_menu is not None:
                self._container_manager.add_to_plugin_menu(self._plugin_menu)

    def _discover(self):
        discovery_data = self._settings.get_settings('discovery_data')
        # check timestamp of the cached discovery data
        cache_stamp = self._settings.value('discovery_timestamp', default_value=None)
        if cache_stamp:
            cache_stamp = float(cache_stamp)

        # wipe cached data when forcing discovery or when cache is to old (or from the future)
        now = time.time()
        if self._application_context.options.force_discover or \
                not cache_stamp or \
                cache_stamp > now or \
                cache_stamp + PluginManager.discovery_cache_max_age < now:
            qDebug('PluginManager._discover() force discovery of plugins')
            cache_stamp = None
            for k in discovery_data.all_keys():
                discovery_data.remove(k)
        else:
            qDebug('PluginManager._discover() using cached plugin discovery information')

        plugin_descriptors = self._plugin_provider.discover(discovery_data)

        # store timestamp and newly discovered data
        if not cache_stamp:
            self._settings.set_value('discovery_timestamp', now)

        return plugin_descriptors

    def find_plugins_by_name(self, lookup_name):
        plugins = {}
        for plugin_id, plugin_full_name in self.get_plugins().items():
            if plugin_full_name.lower().find(lookup_name.lower()) >= 0 or \
                    plugin_id.lower().find(lookup_name.lower()) >= 0:
                plugins[plugin_id] = plugin_full_name
        return plugins

    def get_plugins(self):
        self.discover()
        plugins = {}
        for plugin_id, plugin_descriptor in self._plugin_descriptors.items():
            plugins[plugin_id] = '/'.join(
                plugin_descriptor.attributes().get('class_type', 'unknown').split('::'))
        return plugins

    def get_plugin_descriptor(self, plugin_id):
        return self._plugin_descriptors.get(plugin_id, None)

    def is_plugin_running(self, plugin_id, serial_number):
        instance_id = PluginInstanceId(plugin_id, serial_number)
        return str(instance_id) in self._running_plugins

    @Slot(str)
    @Slot(str, int)
    def load_plugin(self, plugin_id, serial_number=None, argv=None):
        qDebug('PluginManager.load_plugin(%s, %s)' %
               (str(plugin_id), str(serial_number) if serial_number is not None else ''))
        # save state of top-level widgets
        self.plugins_about_to_change_signal.emit()
        if serial_number is None:
            serial_number = self._next_serial_number(plugin_id)
        instance_id = PluginInstanceId(plugin_id, serial_number)
        self._load_plugin_load(instance_id, self._load_plugin_restore, argv)

    def _next_serial_number(self, plugin_id):
        # convert from unicode
        plugin_id = str(plugin_id)
        # collect serial numbers of all running instances of the specific plugin
        used_serial_numbers = {}
        for info in self._running_plugins.values():
            if info['instance_id'].plugin_id == plugin_id:
                used_serial_numbers[info['instance_id'].serial_number] = None

        # find first not used serial number
        serial_number = 1
        while serial_number in used_serial_numbers:
            serial_number = serial_number + 1
        return serial_number

    def _load_plugin_load(self, instance_id, callback, argv=None):
        # if the requested instance is already running, do nothing
        if str(instance_id) in self._running_plugins:
            raise Exception(
                'PluginManager._load_plugin(%s) instance already loaded' % str(instance_id))

        # containers are pseudo-plugins and handled by a special handler
        if self._container_manager is not None and \
            instance_id.plugin_id == \
                self._container_manager.get_container_descriptor().plugin_id():
            handler = PluginHandlerContainer(self, self._main_window, instance_id,
                                             self._application_context, self._container_manager)

        # use platform specific handler for multiprocess-mode if available
        elif self._application_context.options.multi_process or \
                self._application_context.options.embed_plugin:
            try:
                from qt_gui.plugin_handler_xembed import PluginHandlerXEmbed
                handler = \
                    PluginHandlerXEmbed(self, self._main_window, instance_id,
                                        self._application_context, self._container_manager, argv)
            except ImportError:
                qCritical(
                    'PluginManager._load_plugin() could not load plugin in a separate process')
                return

        # use direct handler for in-process plugins
        else:
            handler = PluginHandlerDirect(self, self._main_window, instance_id,
                                          self._application_context, self._container_manager, argv)

        handler.set_minimized_dock_widgets_toolbar(self._minimized_dock_widgets_toolbar)

        if instance_id.plugin_id not in self._plugin_descriptors.keys():
            qWarning(
                'PluginManager._load_plugin() could not load plugin "%s": plugin not available' %
                (instance_id.plugin_id))
            return
        plugin_descriptor = self._plugin_descriptors[instance_id.plugin_id]
        handler.set_plugin_descriptor(plugin_descriptor)

        self._add_running_plugin(instance_id, handler)
        handler.load(self._plugin_provider, callback)

    def _add_running_plugin(self, instance_id, handler):
        if self._plugin_menu is not None:
            plugin_descriptor = self._plugin_descriptors[instance_id.plugin_id]
            self._plugin_menu.add_instance(plugin_descriptor, instance_id)
            handler.label_updated.connect(self._plugin_menu.update_plugin_instance_label)

        info = {
            'handler': handler,
            'instance_id': instance_id
        }
        self._running_plugins[str(instance_id)] = info

    def _load_plugin_restore(self, handler, exception):
        qDebug('PluginManager._load_plugin_restore()')
        self._load_plugin_completed(handler, exception)
        if exception is None:
            # restore settings after load
            self._restore_plugin_settings(handler.instance_id(), self._emit_load_plugin_completed)

    def _load_plugin_completed(self, handler, exception):
        instance_id = handler.instance_id()
        if exception is not None:
            if isinstance(exception, PluginLoadError):
                qWarning('PluginManager._load_plugin() could not load plugin "%s": %s' %
                         (instance_id.plugin_id, exception))
            else:
                qCritical('PluginManager._load_plugin() could not load plugin "%s"%s' %
                          (instance_id.plugin_id, (':\n%s' % traceback.format_exc() if
                           isinstance(exception, Exception) else '')))
            self._remove_running_plugin(instance_id)
            # quit embed application
            if self._application_context.options.embed_plugin:
                exit(-1)
            return

        qDebug('PluginManager._load_plugin(%s) successful' % str(instance_id))

        handler.close_signal.connect(self.unload_plugin)
        handler.reload_signal.connect(self.reload_plugin)
        handler.help_signal.connect(self._emit_plugin_help_signal)
        self.plugin_load_unload_signal.emit(len(self._running_plugins) == 0)

    def _emit_plugin_help_signal(self, instance_id_str):
        instance_id = PluginInstanceId(instance_id=instance_id_str)
        plugin_descriptor = self._plugin_descriptors[instance_id.plugin_id]
        self.plugin_help_signal.emit(plugin_descriptor)

    def _restore_plugin_settings(self, instance_id, callback):
        if self._global_settings is not None and self._perspective_settings is not None:
            info = self._running_plugins[str(instance_id)]
            plugin_settings = self._global_settings.get_settings(
                'plugin__' + instance_id.tidy_plugin_str())
            instance_settings = self._perspective_settings.get_settings(
                'plugin__' + instance_id.tidy_str())
            handler = info['handler']
            handler.restore_settings(plugin_settings, instance_settings, callback)
        else:
            callback(instance_id)

    def _emit_load_plugin_completed(self, instance_id):
        qDebug('PluginManager._emit_load_plugin_completed()')
        # restore state of top-level widgets
        self.plugins_changed_signal.emit()

    @Slot(str)
    def unload_plugin(self, instance_id_str):
        # unloading a plugin with locked perspective or running standalone
        # triggers close of application
        if self._application_context.options.lock_perspective or \
                self._application_context.options.standalone_plugin:
            self._close_application_signal()
            return
        instance_id = PluginInstanceId(instance_id=instance_id_str)
        qDebug('PluginManager.unload_plugin(%s)' % str(instance_id))
        # save state of top-level widgets
        self.plugins_about_to_change_signal.emit()
        # save settings before shutdown and unloading
        self._save_plugin_settings(instance_id, self._unload_plugin_shutdown)

    def _save_plugin_settings(self, instance_id, callback):
        if self._global_settings is not None and self._perspective_settings is not None:
            info = self._running_plugins[str(instance_id)]
            plugin_settings = self._global_settings.get_settings(
                'plugin__' + instance_id.tidy_plugin_str())
            instance_settings = self._perspective_settings.get_settings(
                'plugin__' + instance_id.tidy_str())
            handler = info['handler']
            handler.save_settings(plugin_settings, instance_settings, callback)
        else:
            callback(instance_id)

    def _unload_plugin_shutdown(self, instance_id):
        qDebug('PluginManager._unload_plugin_shutdown(%s)' % str(instance_id))
        self._shutdown_plugin(instance_id, self._unload_plugin_unload)

    def _shutdown_plugin(self, instance_id, callback):
        # shutdown plugin before unloading
        info = self._running_plugins[str(instance_id)]
        handler = info['handler']
        handler.close_signal.disconnect(self.unload_plugin)
        handler.shutdown_plugin(callback)

    def _unload_plugin_unload(self, instance_id):
        qDebug('PluginManager._unload_plugin_unload(%s)' % str(instance_id))
        self._unload_plugin(instance_id, self._unload_plugin_completed)

    def _unload_plugin(self, instance_id, callback=None):
        # unload plugin
        info = self._running_plugins[str(instance_id)]
        handler = info['handler']
        handler.unload(callback)

    def _unload_plugin_completed(self, instance_id):
        qDebug('PluginManager._unload_plugin_completed(%s)' % str(instance_id))
        self._remove_running_plugin(instance_id)
        self.plugin_load_unload_signal.emit(len(self._running_plugins) == 0)

    def _remove_running_plugin(self, instance_id):
        info = self._running_plugins[str(instance_id)]
        if self._plugin_menu is not None:
            self._plugin_menu.remove_instance(instance_id)
            info['handler'].label_updated.disconnect(
                self._plugin_menu.update_plugin_instance_label)
        self._running_plugins.pop(str(instance_id))

    @Slot(str)
    def reload_plugin(self, instance_id_str):
        instance_id = PluginInstanceId(instance_id=instance_id_str)
        qDebug('PluginManager.reload_plugin(%s)' % str(instance_id))
        # save state of top-level widgets
        self.plugins_about_to_change_signal.emit()
        self._reload_plugin_save(instance_id)

    def _reload_plugin_save(self, instance_id):
        # save settings before unloading
        self._save_plugin_settings(instance_id, self._reload_plugin_shutdown)

    def _reload_plugin_shutdown(self, instance_id):
        qDebug('PluginManager._reload_plugin_shutdown(%s)' % str(instance_id))
        self._shutdown_plugin(instance_id, self._reload_plugin_unload)

    def _reload_plugin_unload(self, instance_id):
        qDebug('PluginManager._reload_plugin_unload(%s)' % str(instance_id))
        self._unload_plugin(instance_id, self._reload_plugin_schedule_load)

    def _reload_plugin_schedule_load(self, instance_id):
        qDebug('PluginManager._reload_plugin_schedule_load(%s)' % str(instance_id))
        self._remove_running_plugin(instance_id)
        self._deferred_reload_plugin_signal.emit(str(instance_id))

    def _reload_plugin_load(self, instance_id_str):
        instance_id = PluginInstanceId(instance_id=instance_id_str)
        qDebug('PluginManager._reload_plugin_load(%s)' % str(instance_id))
        self._load_plugin_load(instance_id, self._reload_plugin_restore)

    def _reload_plugin_restore(self, handler, exception):
        qDebug('PluginManager._reload_plugin_restore()')
        self._load_plugin_completed(handler, exception)
        if exception is None:
            # restore settings after load
            self._restore_plugin_settings(handler.instance_id(), self._emit_load_plugin_completed)

    def save_settings(self, global_settings, perspective_settings):
        self._save_settings(global_settings, perspective_settings, self._save_settings_callback)

    def _save_settings(self, global_settings, perspective_settings, callback):
        qDebug('PluginManager.save_settings()')
        self._global_settings = global_settings.get_settings('pluginmanager')
        self._perspective_settings = perspective_settings.get_settings('pluginmanager')
        self._store_running_plugins()
        # trigger async call on all running plugins
        self._number_of_ongoing_calls = len(self._running_plugins)
        if self._number_of_ongoing_calls > 0:
            # If changing perspectives, plugins may be removed from this dictionary during
            # _save_plugin_settings, so a shallow copy of the values is needed
            for info in list(self._running_plugins.values()):
                self._save_plugin_settings(info['instance_id'], callback)
        else:
            callback()

    def _store_running_plugins(self):
        if self._perspective_settings is not None:
            plugins = {}
            for info in self._running_plugins.values():
                instance_id = info['instance_id']
                plugin_id = instance_id.plugin_id
                if plugin_id not in plugins:
                    plugins[plugin_id] = []
                plugins[plugin_id].append(instance_id.serial_number)
            self._perspective_settings.set_value('running-plugins', plugins)

    def _save_settings_callback(self, instance_id=None):
        if instance_id is not None:
            self._number_of_ongoing_calls = self._number_of_ongoing_calls - 1
        if self._number_of_ongoing_calls == 0:
            qDebug('PluginManager.save_settings() completed')
            self._number_of_ongoing_calls = None
            self.save_settings_completed_signal.emit()

    def close_application(self, global_settings, perspective_settings):
        self._save_settings(
            global_settings, perspective_settings, self._close_application_save_callback)

    def _close_application_save_callback(self, instance_id=None):
        self._save_settings_callback(instance_id)
        if self._number_of_ongoing_calls is None:
            self._close_application_shutdown_plugins()

    def _close_application_shutdown_plugins(self):
        # trigger async call on all running plugins
        self._number_of_ongoing_calls = len(self._running_plugins)
        if self._number_of_ongoing_calls > 0:
            for info in self._running_plugins.values():
                self._shutdown_plugin(
                    info['instance_id'], self._close_application_shutdown_callback)
        else:
            self._close_application_shutdown_callback()

    def _close_application_shutdown_callback(self, instance_id=None):
        if instance_id is not None:
            self._number_of_ongoing_calls = self._number_of_ongoing_calls - 1
        if self._number_of_ongoing_calls == 0:
            qDebug('PluginManager.close_application() completed')
            self._number_of_ongoing_calls = None
            self._close_application_signal()

    def _close_application_signal(self):
        self._plugin_provider.shutdown()
        self.close_application_signal.emit()

    def restore_settings(self, global_settings, perspective_settings):
        qDebug('PluginManager.restore_settings()')
        self._global_settings = global_settings.get_settings('pluginmanager')
        self._perspective_settings = perspective_settings.get_settings('pluginmanager')
        self._restore_settings_save_obsolete()

    def _restore_settings_save_obsolete(self):
        # trigger shutdown of obsolete plugins
        plugins = self._restore_running_plugins_get_plugins()
        obsolete = []
        for instance_id in self._running_plugins.keys():
            if instance_id not in plugins.keys():
                obsolete.append(PluginInstanceId(instance_id=instance_id))
        self._number_of_ongoing_calls = len(obsolete)
        if self._number_of_ongoing_calls > 0:
            qDebug('PluginManager.restore_settings() unloading %d obsolete plugins' %
                   self._number_of_ongoing_calls)
            for instance_id in obsolete:
                self._shutdown_plugin(instance_id, self._restore_settings_unload_obsolete)
        else:
            self._restore_settings_unload_obsolete_callback()

    def _restore_running_plugins_get_plugins(self):
        plugins = {}
        if self._perspective_settings is not None:
            data = self._perspective_settings.value('running-plugins', {})
            for plugin_id, serial_numbers in data.items():
                for serial_number in serial_numbers:
                    instance_id = PluginInstanceId(plugin_id, serial_number)
                    plugins[str(instance_id)] = instance_id
        return plugins

    def _restore_settings_unload_obsolete(self, instance_id):
        # trigger unload of obsolete plugins
        self._unload_plugin(instance_id, self._restore_settings_unload_obsolete_callback)

    def _restore_settings_unload_obsolete_callback(self, instance_id=None):
        if instance_id is not None:
            self._number_of_ongoing_calls = self._number_of_ongoing_calls - 1
            self._remove_running_plugin(instance_id)
        if self._number_of_ongoing_calls == 0:
            if instance_id is not None:
                qDebug('PluginManager.restore_settings() all obsolete plugins unloaded')
            self._number_of_ongoing_calls = None
            self._restore_settings_load_missing()

    def _restore_settings_load_missing(self):
        # trigger_load of not yet loaded plugins
        plugins = self._restore_running_plugins_get_plugins()
        loading = []
        for instance_id_str, instance_id in plugins.items():
            if instance_id_str not in self._running_plugins.keys():
                loading.append(instance_id)
        self._number_of_ongoing_calls = len(loading)
        if self._number_of_ongoing_calls > 0:
            qDebug('PluginManager.restore_settings() loading %d plugins' %
                   self._number_of_ongoing_calls)
            for instance_id in loading:
                self._load_plugin_load(instance_id, self._restore_settings_load_missing_callback)
        else:
            self._restore_settings_load_missing_callback()

    def _restore_settings_load_missing_callback(self, handler=None, exception=None):
        if handler is not None:
            self._number_of_ongoing_calls = self._number_of_ongoing_calls - 1
            self._load_plugin_completed(handler, exception)
        if self._number_of_ongoing_calls == 0:
            if handler is not None:
                qDebug('PluginManager.restore_settings() all missing plugins loaded')
            self._number_of_ongoing_calls = None
            self._restore_settings_restore()

    def restore_settings_without_plugins(self, global_settings, perspective_settings):
        qDebug('PluginManager.restore_settings_without_plugins()')
        self._global_settings = global_settings.get_settings('pluginmanager')
        self._perspective_settings = perspective_settings.get_settings('pluginmanager')
        self._restore_settings_restore()

    def _restore_settings_restore(self):
        # trigger restore settings for all running plugins
        self._number_of_ongoing_calls = len(self._running_plugins)
        if self._number_of_ongoing_calls > 0:
            for info in self._running_plugins.values():
                self._restore_plugin_settings(
                    info['instance_id'], self._restore_settings_restore_callback)
        else:
            self._restore_settings_restore_callback()

    def _restore_settings_restore_callback(self, instance_id=None):
        if instance_id is not None:
            self._number_of_ongoing_calls = self._number_of_ongoing_calls - 1
        if self._number_of_ongoing_calls == 0:
            if instance_id is not None:
                qDebug('PluginManager.restore_settings() all plugin settings restored')
            self._number_of_ongoing_calls = None
            # restore state of top-level widgets
            self.plugins_changed_signal.emit()



# Copyright (c) 2011, Dirk Thomas, Dorian Scholz, TU Darmstadt
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above
#     copyright notice, this list of conditions and the following
#     disclaimer in the documentation and/or other materials provided
#     with the distribution.
#   * Neither the name of the TU Darmstadt nor the names of its
#     contributors may be used to endorse or promote products derived
#     from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from python_qt_binding.QtCore import QObject


class PluginProvider(QObject):
    """Interface for discovering, loading and unloading of plugins."""

    def __init__(self):
        super(PluginProvider, self).__init__()
        self.setObjectName('PluginProvider')

    def discover(self, discovery_data):
        """
        Discover the plugins.

        @param discovery_data:
            The settings containing any discovery data which is cached between invocations
        @return: Dictionary of plugin ids to `PluginDescriptor`s
        """
        raise NotImplementedError

    def load(self, plugin_id, plugin_context):
        """
        Load a plugin and pass the `PluginContext`.

        @return: The created plugin instance
        @raise PluginLoadError
        """
        raise NotImplementedError

    def unload(self, plugin_instance):
        """Unload a plugin."""
        raise NotImplementedError

    def shutdown(self):
        """Notify of shutdown."""
        pass



# Copyright (c) 2011, Dirk Thomas, Dorian Scholz, TU Darmstadt
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above
#     copyright notice, this list of conditions and the following
#     disclaimer in the documentation and/or other materials provided
#     with the distribution.
#   * Neither the name of the TU Darmstadt nor the names of its
#     contributors may be used to endorse or promote products derived
#     from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.


class ApplicationContext(object):
    """Application wide context containing variables used in various places."""

    def __init__(self):
        self.provide_app_dbus_interfaces = False
        """The flag if the application should listen to commands on DBus."""

        self.dbus_base_bus_name = None
        """The named DBus bus name of the application."""

        self.dbus_unique_bus_name = None
        """The unique DBus bus name of the application."""

        self.dbus_host_bus_name = None
        """The unique DBus bus name of another application to communicate with."""

        self.options = None
        """The parsed command line options."""

        self.qtgui_path = None
        """The path of the qt_gui package."""



# Copyright (c) 2011, Dirk Thomas, Dorian Scholz, TU Darmstadt
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above
#     copyright notice, this list of conditions and the following
#     disclaimer in the documentation and/or other materials provided
#     with the distribution.
#   * Neither the name of the TU Darmstadt nor the names of its
#     contributors may be used to endorse or promote products derived
#     from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import traceback

from python_qt_binding.QtCore import qCritical, qDebug, QObject, Qt, qWarning, Signal, Slot
from python_qt_binding.QtWidgets import QDockWidget, QToolBar

from qt_gui.dock_widget import DockWidget
from qt_gui.dock_widget_title_bar import DockWidgetTitleBar
from qt_gui.icon_loader import get_icon
from qt_gui.window_changed_signaler import WindowChangedSignaler


class PluginHandler(QObject):
    """
    Base class for the bidirectional exchange between the framework and one `Plugin` instance.

    It utilizes a `PluginProvider` to load/unload the plugin and provides callbacks for the
    `PluginContext`.
    """

    label_updated = Signal(str, str)
    close_signal = Signal(str)
    reload_signal = Signal(str)
    help_signal = Signal(str)
    _defered_check_close = Signal()

    def __init__(self, parent, main_window, instance_id, application_context, container_manager,
                 argv=None):
        super(PluginHandler, self).__init__(parent)
        self.setObjectName('PluginHandler')

        self._main_window = main_window
        self._instance_id = instance_id
        self._application_context = application_context
        self._container_manager = container_manager
        self._argv = argv if argv else []
        self._minimized_dock_widgets_toolbar = None
        self._plugin_descriptor = None

        self._defered_check_close.connect(self._check_close, Qt.QueuedConnection)
        self._plugin_provider = None
        self.__callback = None
        self.__instance_settings = None

        self._plugin_has_configuration = False

        # mapping of added widgets to their parent dock widget and WindowChangedSignaler
        self._widgets = {}

        self._toolbars = []

    def instance_id(self):
        return self._instance_id

    def argv(self):
        return self._argv

    def set_minimized_dock_widgets_toolbar(self, toolbar):
        self._minimized_dock_widgets_toolbar = toolbar

    def set_plugin_descriptor(self, plugin_descriptor):
        self._plugin_descriptor = plugin_descriptor

    def load(self, plugin_provider, callback=None):
        """
        Load plugin.

        Completion is signaled asynchronously if a callback is passed.
        """
        self._plugin_provider = plugin_provider
        self.__callback = callback
        try:
            self._load()
        except Exception as e:
            self._emit_load_completed(e)

    def _load(self):
        raise NotImplementedError

    def _emit_load_completed(self, exception=None):
        if exception is not None:
            self._garbage_widgets_and_toolbars()
        if self.__callback is not None:
            callback = self.__callback
            self.__callback = None
            callback(self, exception)
        elif exception is not None:
            qCritical('PluginHandler.load() failed%s' %
                      (':\n%s' % str(exception) if not exception else ''))

    def _garbage_widgets_and_toolbars(self):
        for widget in list(self._widgets.keys()):
            self.remove_widget(widget)
            self._delete_widget(widget)
        for toolbar in list(self._toolbars):
            self.remove_toolbar(toolbar)
            self._delete_toolbar(toolbar)

    def shutdown_plugin(self, callback):
        """
        Shut down the plugin and remove all added widgets.

        Completion is signaled asynchronously if a callback is passed.
        """
        self.__callback = callback
        try:
            self._shutdown_plugin()
        except Exception:
            qCritical('PluginHandler.shutdown_plugin() plugin "%s" raised an exception:\n%s' %
                      (str(self._instance_id), traceback.format_exc()))
            self.emit_shutdown_plugin_completed()

    def _shutdown_plugin(self):
        raise NotImplementedError

    def emit_shutdown_plugin_completed(self):
        self._garbage_widgets_and_toolbars()
        if self.__callback is not None:
            callback = self.__callback
            self.__callback = None
            callback(self._instance_id)

    def _delete_widget(self, widget):
        widget.deleteLater()

    def _delete_toolbar(self, toolbar):
        toolbar.deleteLater()

    def unload(self, callback=None):
        """
        Unload plugin.

        Completion is signaled asynchronously if a callback is passed.
        """
        self.__callback = callback
        try:
            self._unload()
        except Exception:
            qCritical('PluginHandler.unload() plugin "%s" raised an exception:\n%s' %
                      (str(self._instance_id), traceback.format_exc()))
            self._emit_unload_completed()

    def _unload(self):
        raise NotImplementedError

    def _emit_unload_completed(self):
        if self.__callback is not None:
            callback = self.__callback
            self.__callback = None
            callback(self._instance_id)

    def save_settings(self, plugin_settings, instance_settings, callback=None):
        """
        Save settings of the plugin and all dock widget title bars.

        Completion is signaled asynchronously if a callback is passed.
        """
        qDebug('PluginHandler.save_settings()')
        self.__instance_settings = instance_settings
        self.__callback = callback
        try:
            self._save_settings(plugin_settings, instance_settings)
        except Exception:
            qCritical('PluginHandler.save_settings() plugin "%s" raised an exception:\n%s' %
                      (str(self._instance_id), traceback.format_exc()))
            self.emit_save_settings_completed()

    def _save_settings(self, plugin_settings, instance_settings):
        raise NotImplementedError

    def emit_save_settings_completed(self):
        qDebug('PluginHandler.emit_save_settings_completed()')
        self._call_method_on_all_dock_widgets('save_settings', self.__instance_settings)
        self.__instance_settings = None
        if self.__callback is not None:
            callback = self.__callback
            self.__callback = None
            callback(self._instance_id)

    def _call_method_on_all_dock_widgets(self, method_name, instance_settings):
        for dock_widget, _, _ in self._widgets.values():
            name = 'dock_widget' + \
                dock_widget.objectName().replace(self._instance_id.tidy_str(), '', 1)
            settings = instance_settings.get_settings(name)
            method = getattr(dock_widget, method_name)
            try:
                method(settings)
            except Exception:
                qCritical('PluginHandler._call_method_on_all_dock_widgets(%s) failed:\n%s' %
                          (method_name, traceback.format_exc()))

    def restore_settings(self, plugin_settings, instance_settings, callback=None):
        """
        Restore settings of the plugin and all dock widget title bars.

        Completion is signaled asynchronously if a callback is passed.
        """
        qDebug('PluginHandler.restore_settings()')
        self.__instance_settings = instance_settings
        self.__callback = callback
        try:
            self._restore_settings(plugin_settings, instance_settings)
        except Exception:
            qCritical('PluginHandler.restore_settings() plugin "%s" raised an exception:\n%s' %
                      (str(self._instance_id), traceback.format_exc()))
            self.emit_restore_settings_completed()

    def _restore_settings(self, plugin_settings, instance_settings):
        raise NotImplementedError

    def emit_restore_settings_completed(self):
        qDebug('PluginHandler.emit_restore_settings_completed()')
        # call after plugin has restored settings as it may spawn additional dock widgets
        self._call_method_on_all_dock_widgets('restore_settings', self.__instance_settings)
        self.__instance_settings = None
        if self.__callback is not None:
            callback = self.__callback
            self.__callback = None
            callback(self._instance_id)

    def _create_dock_widget(self):
        dock_widget = DockWidget(self._container_manager)
        self._update_dock_widget_features(dock_widget)
        self._update_title_bar(dock_widget)
        self._set_window_icon(dock_widget)
        return dock_widget

    def _update_dock_widget_features(self, dock_widget):
        if self._application_context.options.lock_perspective or \
                self._application_context.options.standalone_plugin:
            # dock widgets are not closable when perspective is locked or plugin is
            # running standalone
            features = dock_widget.features()
            dock_widget.setFeatures(features ^ QDockWidget.DockWidgetClosable)
        if self._application_context.options.freeze_layout:
            # dock widgets are not closable when perspective is locked or plugin is
            # running standalone
            features = dock_widget.features()
            dock_widget.setFeatures(
                features ^ (QDockWidget.DockWidgetMovable | QDockWidget.DockWidgetFloatable))

    def _update_title_bar(self, dock_widget, hide_help=False, hide_reload=False):
        title_bar = dock_widget.titleBarWidget()
        if title_bar is None:
            title_bar = DockWidgetTitleBar(
                dock_widget, self._application_context.qtgui_path,
                hide_title=self._application_context.options.hide_title)
            dock_widget.setTitleBarWidget(title_bar)

            # connect extra buttons
            title_bar.connect_close_button(self._remove_widget_by_dock_widget)
            title_bar.connect_button('help', self._emit_help_signal)
            if hide_help:
                title_bar.show_button('help', not hide_help)
            title_bar.connect_button('reload', self._emit_reload_signal)
            if hide_reload:
                title_bar.show_button('reload', not hide_reload)
            title_bar.connect_button('configuration', self._trigger_configuration)
            title_bar.show_button('configuration', self._plugin_has_configuration)

    def _set_window_icon(self, widget):
        if self._plugin_descriptor:
            action_attributes = self._plugin_descriptor.action_attributes()
            if 'icon' in action_attributes and action_attributes['icon'] is not None:
                base_path = self._plugin_descriptor.attributes().get('plugin_path')
                icon = get_icon(
                    action_attributes['icon'], action_attributes.get('icontype', None), base_path)
                widget.setWindowIcon(icon)

    def _update_title_bars(self):
        if self._plugin_has_configuration:
            for dock_widget, _, _ in self._widgets.values():
                title_bar = dock_widget.titleBarWidget()
                title_bar.show_button('configuration')

    def _remove_widget_by_dock_widget(self, dock_widget):
        widget = [key for key, value in self._widgets.items() if value[0] == dock_widget][0]
        self.remove_widget(widget)

    def _emit_help_signal(self):
        self.help_signal.emit(str(self._instance_id))

    def _emit_reload_signal(self):
        self.reload_signal.emit(str(self._instance_id))

    def _trigger_configuration(self):
        self._plugin.trigger_configuration()

    def _add_dock_widget(self, dock_widget, widget):
        dock_widget.setWidget(widget)
        # every dock widget needs a unique name for save/restore geometry/state to work
        dock_widget.setObjectName(self._instance_id.tidy_str() + '__' + widget.objectName())
        self._add_dock_widget_to_main_window(dock_widget)
        signaler = WindowChangedSignaler(widget, widget)
        signaler.window_icon_changed_signal.connect(self._on_widget_icon_changed)
        signaler.window_title_changed_signal.connect(self._on_widget_title_changed)
        signaler2 = WindowChangedSignaler(dock_widget, dock_widget)
        signaler2.hide_signal.connect(self._on_dock_widget_hide)
        signaler2.show_signal.connect(self._on_dock_widget_show)
        signaler2.window_title_changed_signal.connect(self._on_dock_widget_title_changed)
        self._widgets[widget] = [dock_widget, signaler, signaler2]
        # trigger to update initial window icon and title
        signaler.emit_all()
        # trigger to update initial window state
        signaler2.emit_all()

    def _add_dock_widget_to_main_window(self, dock_widget):
        if self._main_window is not None:
            # warn about dock_widget with same object name
            old_dock_widget = self._main_window.findChild(DockWidget, dock_widget.objectName())
            if old_dock_widget is not None:
                qWarning('PluginHandler._add_dock_widget_to_main_window() duplicate object name ' +
                         '"%s", assign unique object names before adding widgets!' %
                         dock_widget.objectName())

            self._main_window.addDockWidget(Qt.BottomDockWidgetArea, dock_widget)

    def _on_widget_icon_changed(self, widget):
        dock_widget, _, _ = self._widgets[widget]
        dock_widget.setWindowIcon(widget.windowIcon())

    def _on_widget_title_changed(self, widget):
        dock_widget, _, _ = self._widgets[widget]
        dock_widget.setWindowTitle(widget.windowTitle())

    def _on_dock_widget_hide(self, dock_widget):
        if self._minimized_dock_widgets_toolbar:
            self._minimized_dock_widgets_toolbar.addDockWidget(dock_widget)

    def _on_dock_widget_show(self, dock_widget):
        if self._minimized_dock_widgets_toolbar:
            self._minimized_dock_widgets_toolbar.removeDockWidget(dock_widget)

    def _on_dock_widget_title_changed(self, dock_widget):
        self.label_updated.emit(str(self._instance_id), dock_widget.windowTitle())

    # pointer to QWidget must be used for PySide to work (at least with 1.0.1)
    @Slot('QWidget*')
    def remove_widget(self, widget):
        dock_widget, signaler, signaler2 = self._widgets[widget]
        self._widgets.pop(widget)
        if signaler is not None:
            signaler.window_icon_changed_signal.disconnect(self._on_widget_icon_changed)
            signaler.window_title_changed_signal.disconnect(self._on_widget_title_changed)
        if signaler2 is not None:
            # emit show signal to remove dock widget from minimized toolbar before removal
            signaler2.show_signal.emit(dock_widget)
            signaler2.hide_signal.disconnect(self._on_dock_widget_hide)
            signaler2.show_signal.disconnect(self._on_dock_widget_show)
        # remove dock widget from parent and delete later
        if self._main_window is not None:
            dock_widget.parent().removeDockWidget(dock_widget)
        # do not delete the widget, only the dock widget
        dock_widget.setParent(None)
        widget.setParent(None)
        dock_widget.deleteLater()
        # defer check for last widget closed to give plugin a chance to add
        # another widget right away
        self._defered_check_close.emit()

    def _add_toolbar(self, toolbar):
        # every toolbar needs a unique name for save/restore geometry/state to work
        toolbar_object_name = toolbar.objectName()
        prefix = self._instance_id.tidy_str() + '__'
        # when added, removed and readded the prefix should not be prepended multiple times
        if not toolbar_object_name.startswith(prefix):
            toolbar_object_name = prefix + toolbar_object_name
        toolbar.setObjectName(toolbar_object_name)

        if self._application_context.options.freeze_layout:
            toolbar.setMovable(False)

        self._toolbars.append(toolbar)
        if self._main_window is not None:
            # warn about toolbar with same object name
            old_toolbar = self._main_window.findChild(QToolBar, toolbar.objectName())
            if old_toolbar is not None:
                qWarning('PluginHandler._add_toolbar() duplicate object name "%s", '
                         'assign unique object names before adding toolbars!' %
                         toolbar.objectName())
            self._main_window.addToolBar(Qt.TopToolBarArea, toolbar)

    # pointer to QToolBar must be used for PySide to work (at least with 1.0.1)
    @Slot('QToolBar*')
    def remove_toolbar(self, toolbar):
        self._toolbars.remove(toolbar)
        # detach toolbar from parent
        if toolbar.parent():
            toolbar.parent().removeToolBar(toolbar)
        # defer check for last widget closed to give plugin a chance to add
        # another widget right away
        self._defered_check_close.emit()

    def _check_close(self):
        # close plugin when no widgets or toolbars are left
        if len(self._widgets) + len(self._toolbars) == 0:
            self._emit_close_plugin()

    def _emit_close_plugin(self):
        self.close_signal.emit(str(self._instance_id))
